/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov [committer] */

/// \file   CbmQaUtil.cxx
/// \brief  Useful utilities for CBM QA tasks
/// \author S.Gorbunov
/// \data   31.07.2023


#include "CbmQaUtil.h"

#include "CbmQaCanvas.h"

#include "TAxis.h"
#include "TCanvas.h"
#include "TF1.h"
#include "TH1.h"
#include "TPaveStats.h"
#include "TVirtualPad.h"

namespace cbm::qa::util
{

  // ---------------------------------------------------------------------------------------------------------------------
  //
  std::tuple<double, double> FitKaniadakisGaussian(TH1* pHist)
  {
    //
    // Fit function - Kaniadakis Gaussian Distribution:
    // https://en.wikipedia.org/wiki/Kaniadakis_Gaussian_distribution
    //
    // where exp_k(x) is calculated via arcsinh():
    // https://en.wikipedia.org/wiki/Kaniadakis_statistics
    //
    // parameters:
    //
    // [0] - mean
    // [1] - Std. Dev ( calculated after the fit )
    // [2] - peak
    // [3] - hwhm - half width at half maximum, scaled.  ( == standard deviation for a gaussian )
    // [4] - k - Kaniadakis parameter, k in [0.,1.]. ( == 0.0 for a gaussian. The formula below breaks when k is exactly 0.)
    //
    // returns mean and std.dev
    //

    double xMin = pHist->GetXaxis()->GetXmin();
    double xMax = pHist->GetXaxis()->GetXmax();

    TF1 fit("FitKaniadakisGaussian", "[2] * TMath::Exp(TMath::ASinH(-0.5*[4]*((x-[0])/[3])**2)/[4]) + 0.0*[1]", xMin,
            xMax, "NL");

    fit.SetParName(0, "fit_Mean");
    fit.SetParName(1, "fit_StdDev");
    fit.SetParName(2, "fit_Peak");
    fit.SetParName(3, "fit_Hwhm");
    fit.SetParName(4, "fit_k");

    // set reasonable initial values

    double mean = pHist->GetMean();
    double peak = pHist->GetBinContent(pHist->GetMaximumBin());
    double hwhm = pHist->GetStdDev();

    fit.SetParameters(mean, hwhm, peak, hwhm, .001);

    // limit parameter k

    fit.SetParLimits(4, 0.00001, 5.);

    // fit in a quite mode

    TVirtualPad* padsav = gPad;
    CbmQaCanvas::GetDummyCanvas().cd();
    pHist->SetStats(0);  // remove the old stat window
    pHist->SetStats(1);  // set the flag to create the stat window during Draw()
    //pHist->Draw();
    //CbmQaCanvas::GetDummyCanvas().Update();

    //pHist->Sumw2();
    if (pHist->GetEntries() > 0) { pHist->Fit(&fit, "Q"); }
    pHist->Draw();
    CbmQaCanvas::GetDummyCanvas().Update();

    TF1* f = pHist->GetFunction("FitKaniadakisGaussian");
    if (nullptr != f) {

      // calculate the Std Dev

      f->SetParameter(1, sqrt(f->CentralMoment(2, xMin, xMax)));
      f->SetParError(1, f->GetParError(3));

      // fix some parameters to prevent them from showing up in the stat window

      f->FixParameter(2, f->GetParameter(2));
      f->FixParameter(3, f->GetParameter(3));
      f->FixParameter(4, f->GetParameter(4));
    }

    TPaveStats* stats = (TPaveStats*) pHist->FindObject("stats");
    assert(stats);

    stats->SetX1NDC(0.7);
    stats->SetY1NDC(0.5);
    stats->SetOptStat(111110);
    //stats->SetOptStat("eMRuo");
    stats->SetOptFit(100001);

    CbmQaCanvas::GetDummyCanvas().Clear();
    if (padsav) padsav->cd();
    return (nullptr != f) ? std::tuple(f->GetParameter(0), f->GetParameter(1)) : std::tuple(0., -1.);
  }

}  // namespace cbm::qa::util
